#define _GNU_SOURCE
#include <stdio.h>
#include <dlfcn.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/types.h>
#include <poll.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/select.h>
#include <string.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <netinet/in.h>

/*
 *  Copyright: Collin Mulliner <collin@mulliner.org>
 * 
 *  Web:       http://www.mulliner.org/security/sms/
 *
 *  License:   GPLv3
 *
 */

#define FD3 "/dev/dlci.spi-baseband.3"
#define FD4 "/dev/dlci.spi-baseband.4"
#define FD3H "/dev/dlci.h5-baseband.3"
#define FD4H "/dev/dlci.h5-baseband.4"
#define SFD3 "/tmp/fuzz3.sock"
#define SFD4 "/tmp/fuzz4.sock"
#define CTL "/tmp/fuzzctl.sock"
#define PORT 4223

#define offsetof(TYPE, MEMBER) ((int)&((TYPE *)0)->MEMBER)

struct state_t {
	int fd3;
	int fd4;

	int ssfd3;
	int csfd3;
	
	int ssfd4;
	int csfd4;

	int sctl;
	int cctl;
	
	int snctl;
	int cnctl;

	int sms_avail;
	int cnma;
	int xtesm;
};

struct state_t s = {0};

static FILE *log;
static int dolog = 1;

static int open_log()
{
	if (dolog == 0)
		return 0;
		
#ifdef TEST
	log = fopen("/tmp/injectord.log", "w");
#else
	log = fopen("/var/root/injectord.log", "w");
#endif	
	if (log == NULL) {
		printf("can't open log file\n");
		exit(0);
	}
	return 1;
}
 
int setup_sock(int *s, char *name)
{
	int len;
	struct sockaddr_un saun;

	*s = socket(AF_UNIX, SOCK_STREAM, 0);	
	saun.sun_family = AF_UNIX;
	strcpy(saun.sun_path, name);
	unlink(name);
	len = offsetof(struct sockaddr_un, sun_path) + strlen(name);
	if (bind(*s, &saun, len) < 0) {
		if (dolog) fprintf(log, "bind failed\n");
		return -1;
	}
	if (listen(*s, 2) < 0) {
		if (dolog) fprintf(log, "listen failed\n");
		return -2;
	}
	return 1;
}

int setup_netsock(int *s, int port)
{
	int len;
	struct sockaddr_in saun ={0};

	*s = socket(AF_INET, SOCK_STREAM, 0);	
	saun.sin_family = AF_INET;
	saun.sin_port = htons(port);
	len = sizeof(saun);
	if (bind(*s, &saun, len) < 0) {
		if (dolog) fprintf(log, "net bind failed\n");
		return -1;
	}
	if (listen(*s, 2) < 0) {
		if (dolog) fprintf(log, "net listen failed\n");
		return -2;
	}
	return 1;
}

int init()
{
	s.fd3 = -1;
	s.fd4 = -1;
	s.ssfd3 = -1;
	s.ssfd4 = -1;
	s.csfd3 = -1;
	s.csfd4 = -1;
	s.sctl = -1;
	s.cctl = -1;
	s.snctl = -1;
	s.cnctl = -1;

	setup_sock(&s.ssfd3, SFD3);
	setup_sock(&s.ssfd4, SFD4);
	setup_sock(&s.sctl, CTL);
	setup_netsock(&s.snctl, PORT);
}

int setup_poll(fd_set *rfds)
{
	int num = 0;

	FD_ZERO(rfds);

	if (s.csfd3 != -1) {
		FD_SET(s.csfd3, rfds);
		if (s.csfd3 > num)
			num = s.csfd3;
	}
	else if (s.ssfd3 != -1) {
		FD_SET(s.ssfd3, rfds);
		if (s.ssfd3 > num)
			num = s.ssfd3;
	}
	if (s.csfd4 != -1) {
		FD_SET(s.csfd4, rfds);
		if (s.csfd4 > num)
			num = s.csfd4;
	}
	else if (s.ssfd4 != -1) {
		FD_SET(s.ssfd4, rfds);
		if (s.ssfd4 > num)
			num = s.ssfd4;
	}
	if (s.cctl != -1) {
		FD_SET(s.cctl, rfds);
		if (s.cctl > num)
			num = s.cctl;
	}
	else if (s.sctl != -1) {
		FD_SET(s.sctl, rfds);
		if (s.sctl > num)
			num = s.sctl;
	}
	if (s.cnctl != -1) {
		FD_SET(s.cnctl, rfds);
		if (s.cnctl > num)
			num = s.cnctl;
	}
	else if (s.snctl != -1) {
		FD_SET(s.snctl, rfds);
		if (s.snctl > num)
			num = s.snctl;
	}
	if (s.fd3 != -1) {
		//fprintf(log, "added fd3 to poll\n");
		FD_SET(s.fd3, rfds);
		if (s.fd3 > num)
			num = s.fd3;
	}
	if (s.fd4 != -1) {
		//fprintf(log, "added fd4 to poll\n");
		FD_SET(s.fd4, rfds);
		if (s.fd4 > num)
			num = s.fd4;
	}
	return num;
}

int main(int argc, char **argv)
{
	struct pollfd pfd[5];

	if (argc > 1)
		dolog = 0;

	open_log();
	init();
		
	for (;;) {
		fflush(log);

		fd_set rfds;
		int numpoll = setup_poll(&rfds);
		int np = select(numpoll + 1, &rfds, NULL, NULL, NULL);
		if (np == -1)
			if (dolog) fprintf(log, "select error\n");

		if (s.csfd3 == -1 && FD_ISSET(s.ssfd3, &rfds)) {
			s.csfd3 = accept(s.ssfd3, NULL, NULL);
			if (s.csfd3 < 0) {
				if (dolog) fprintf(log, "ssfd3 failed\n");
				s.csfd3 = -1;
				if (s.fd3 != -1) {
					if (dolog) fprintf(log, "ssfd3 disconnected\n");
					close(s.fd3);
					s.fd3 = -1;
				}
			}
			else {
				if (dolog) fprintf(log, "ssfd3 connected\n");
				s.fd3 = open(FD3, O_RDWR);
				if (s.fd3 >= 0) {
					if (dolog) fprintf(log, "%s opened\n", FD3);
				}
				else {
					if (dolog) fprintf(log, "%s failed = %d\n", FD3, s.fd3);
				}
				if (s.fd3 == -1) {
					s.fd3 = open(FD3H, O_RDWR);
					if (s.fd3 >= 0) {
						if (dolog) fprintf(log, "%s opened\n", FD3H);
					}
					else {
						if (dolog) fprintf(log, "%s failed = %d\n", FD3H, s.fd3);
					}
				}
			}
		}
		else if (s.csfd3 != -1 && FD_ISSET(s.csfd3, &rfds)) {
			unsigned char buffer[1024*2];
			int rn = read(s.csfd3, buffer, sizeof(buffer));
			if (rn <= 0) {
				close(s.csfd3);
				s.csfd3 = -1;
				if (dolog) fprintf(log, "csfd3 disconnected\n");
			}
			else {
				if (s.sms_avail == 1 && s.xtesm == 0) {
					s.xtesm = 1;
					if (dolog) fprintf(log, "swollow at+xtesm\n");
					int wb = write(s.csfd3, "\nOK\n\n", 5);
					if (s.sms_avail == 1 && s.xtesm == 1 && s.cnma == 1) {
						if (dolog) fprintf(log, "sms xfer done\n");
						s.sms_avail = 0;
						s.cnma = 0;
						s.xtesm = 0;
					}	
				}
				else {
					int wb = write(s.fd3, buffer, rn);
					if (dolog) {
						fprintf(log, "csfd3 to fd3 write %d bytes\n", wb);
						fprintf(log, "---\n");
						int i;
						for (i = 0; i < rn; i++) {
							fprintf(log, "%c", buffer[i]);
						}
						fprintf(log, "\n+++\n");
					}
				}
			}
		}

		if (s.csfd4 == -1 && FD_ISSET(s.ssfd4, &rfds)) {
			s.csfd4 = accept(s.ssfd4, NULL, NULL);
			if (s.csfd4 < 0) {
				if (dolog) fprintf(log, "ssfd4 failed\n");
				s.csfd4 = -1;
				if (s.fd4 != -1) {
					if (dolog) fprintf(log, "ssfd4 disconnected\n");
					close(s.fd4);
					s.fd4 = -1;
				}
			}
			else {
				if (dolog) fprintf(log, "ssfd4 connected\n");
				s.fd4 = open(FD4, O_RDWR);
				if (s.fd4 >= 0) {
					if (dolog) fprintf(log, "%s opened\n", FD4);
				}
				else {
					if (dolog) fprintf(log, "%s failed = %d\n", FD4, s.fd4);
				}
				if (s.fd4 == -1) {
					s.fd4 = open(FD4H, O_RDWR);
					if (s.fd4 >= 0) {
						if (dolog) fprintf(log, "%s opened\n", FD4H);
					}
					else {
						if (dolog) fprintf(log, "%s failed = %d\n", FD4H, s.fd4);
					}
				}
			}
		}
		else if (s.csfd4 != -1 && FD_ISSET(s.csfd4, &rfds)) {
			unsigned char buffer[1024*2];
			int rn = read(s.csfd4, buffer, sizeof(buffer));
			if (rn <= 0) {
				close(s.csfd4);
				s.csfd4 = -1;
				if (dolog) fprintf(log, "csfd4 disconnected\n");
			}
			else {
				if (s.sms_avail == 1 && s.cnma == 0) {
					s.cnma = 1;
					if (dolog) fprintf(log, "swollow at+cnma\n");
					int wb = write(s.csfd4, "\nOK\n\n", 5);
					if (s.sms_avail == 1 && s.xtesm == 1 && s.cnma == 1) {
						if (dolog) fprintf(log, "sms xfer done\n");
						s.sms_avail = 0;
						s.cnma = 0;
						s.xtesm = 0;
					}	
				}
				else {
					int wb = write(s.fd4, buffer, rn);
					if (dolog) {
						fprintf(log, "csfd4 to fd4 write %d bytes\n", wb);
						fprintf(log, "---\n");
						int i;
						for (i = 0; i < rn; i++) {
							fprintf(log, "%c", buffer[i]);
						}
						fprintf(log, "\n+++\n");
					}
				}
			}
		}
		
		if (s.cctl == -1 && FD_ISSET(s.sctl, &rfds)) {
			s.cctl = accept(s.sctl, NULL, NULL);
			if (s.cctl < 0) {
				s.cctl = -1;
				if (dolog) fprintf(log, "fuzz ctl connection error\n");
			}
			else {
				if (dolog) fprintf(log, "fuzz ctl connected\n");
			}
		}
		else if (s.cctl != -1 && FD_ISSET(s.cctl, &rfds)) {
			char buffer[1024*2] = {0};
			int rn = read(s.cctl, buffer, sizeof(buffer));
			if (rn <= 0) {
				close(s.cctl);
				s.cctl = -1;
				if (dolog) fprintf(log, "fuzz ctl disconnected\n");
			}
			else {
				int wr = write(s.csfd3, buffer, rn);
				if (wr <= 0) {
					close(s.csfd3);
					s.csfd3 = -1;
					if (dolog) fprintf(log, "FAILED fuzz command: %s\n", buffer);
				}
				else {
					if (dolog) fprintf(log, "fuzz command: %s\n", buffer);
					s.sms_avail = 1;
					s.cnma = 0;
					s.xtesm = 0;
				}
			}
		}
	
		if (s.cnctl == -1 && FD_ISSET(s.snctl, &rfds)) {
			s.cnctl = accept(s.snctl, NULL, NULL);
			if (s.cnctl < 0) {
				s.cnctl = -1;
				if (dolog) fprintf(log, "fuzz net ctl connection error\n");
			}
			else {
				if (dolog) fprintf(log, "fuzz net ctl connected\n");
			}
		}
		else if (s.cnctl != -1 && FD_ISSET(s.cnctl, &rfds)) {
			char buffer[1024*2] = {0};
			int rn = read(s.cnctl, buffer, sizeof(buffer));
			if (rn <= 0) {
				close(s.cnctl);
				s.cnctl = -1;
				if (dolog) fprintf(log, "fuzz net ctl disconnected\n");
			}
			else {
				int wr = write(s.csfd3, buffer, rn);
				if (wr <= 0) {
					close(s.csfd3);
					s.csfd3 = -1;
					if (dolog) fprintf(log, "FAILED net fuzz command: %s\n", buffer);
				}
				else {
					if (dolog) fprintf(log, "fuzz net command: %s\n", buffer);
					s.sms_avail = 1;
					s.cnma = 0;
					s.xtesm = 0;
				}
			}
		}
	
		if (s.fd3 != -1 && FD_ISSET(s.fd3, &rfds)) {
			char buffer[1024*2] = {0};
			int rn = read(s.fd3, buffer, sizeof(buffer));
			if (rn <= 0) {
				close(s.fd3);
				s.fd3 = -1;
				if (dolog) fprintf(log, "fd3 disconncted (3)\n");
			}
			else {
				int wb = write(s.csfd3, buffer, rn);
				if (wb <= 0) {
					if (dolog) fprintf(log, "csfd3 disconncted (3)\n");
					close(s.csfd3);
					s.csfd3 = -1;
				}
				else {
					if (dolog) {
						fprintf(log, "fd3 to csfd3 write %d bytes\n", wb);
						fprintf(log, "---\n");
						int i;
						for (i = 0; i < rn; i++) {
							fprintf(log, "%c", buffer[i]);
						}
						fprintf(log, "\n+++\n");
					}
				}
			}
		}
		
		if (s.fd4 != -1 && FD_ISSET(s.fd4, &rfds)) {
			char buffer[1024*2] = {0};
			int rn = read(s.fd4, buffer, sizeof(buffer));
			if (rn <= 0) {
				close(s.fd4);
				s.fd4 = -1;
				if (dolog) fprintf(log, "fd4 disconncted (3)\n");
			}
			else {
				int wb = write(s.csfd4, buffer, rn);
				if (wb <= 0) {
					if (dolog) fprintf(log, "csfd4 disconncted (3)\n");
					close(s.csfd4);
					s.csfd4 = -1;
				}
				else {
					if (dolog) {
						fprintf(log, "fd4 to csfd4 write %d bytes\n", wb);
						fprintf(log, "---\n");
						int i;
						for (i = 0; i < rn; i++) {
							fprintf(log, "%c", buffer[i]);
						}
						fprintf(log, "\n+++\n");
					}
				}
			}
		}

	} // for
}
