#define _GNU_SOURCE
#include <stdio.h>
#include <dlfcn.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/types.h>
#include <poll.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/select.h>
#include <string.h>
#include <sys/un.h>
#include <sys/socket.h>

/*
 *  Copyright: Collin Mulliner <collin@mulliner.org>
 * 
 *  Web:       http://www.mulliner.org/security/sms/
 *
 *  License:   GPLv3
 *
 */

#define offsetof(TYPE, MEMBER) ((int)&((TYPE *)0)->MEMBER)

#define FD3 "/tmp/fuzz3.sock"
#define FD4 "/tmp/fuzz4.sock"

static int fd3 = -1;
static int fd4 = -1;

static int log(char *l, unsigned char *bin, unsigned long leng)
{
	// disable logging
	//return 0;

	ssize_t (*real_write) (int fd, const void *buf, size_t count);
	int (*real_open) (const char*, int, ...);
	
	real_write = dlsym(RTLD_NEXT, "write");
	real_open = dlsym(RTLD_NEXT, "open");

#ifdef TEST
	int fp = real_open("/tmp/ldpre.log", O_APPEND|O_RDWR);
#else
	int fp = real_open("/var/root/ldpre.log", O_APPEND|O_RDWR);
#endif
	if (fp) {
		real_write(fp, l, strlen(l));
		if (leng > 0) {
			real_write(fp, "---\n", 4);
			real_write(fp, bin, leng);
			real_write(fp, "+++\n", 4);
		}
		close(fp);
	}
	return 0;
}

ssize_t write(int fd, const void *buf, size_t count)
{
	char l[512] = {0};
	ssize_t (*real_write)(int fd, const void *buf, size_t count);
	
	real_write = dlsym(RTLD_NEXT, "write");
		
	ssize_t res = real_write(fd, buf, count);
	
	if (fd == fd3 || fd == fd4) {
		sprintf(l, "%d = write(%d, x, %d)\n", res, fd, count);
		log(l, buf, res);
	}
	
	return res;
}

ssize_t read(int fd, void *buf, size_t count)
{
	char l[512] = {0};
	ssize_t (*real_read)(int fd, void *buf, size_t count);
	
	real_read = dlsym(RTLD_NEXT, "read");
		
	ssize_t res = real_read(fd, buf, count);
	
	if (fd == fd3 || fd == fd4) {
		sprintf(l, "%d = read(%d, x, %d)\n", res, fd, count);
		log(l, buf, res);
	}
	
	return res;
}

int open(const char *path, int flags, ...)
{
	char l[512] = {0};

	sprintf(l, "entering open(%s) ", path);
	log(l, 0, 0);

	int (*real_open) (const char*, int, ...);
	//printf("entering open(): %s\n", path);
	real_open = dlsym(RTLD_NEXT, "open");

	if (dlerror()) {
		sprintf(l, "open_dlerror ");
		log(l, 0, 0);
		return -1;
	}
  
	int fd = -1;
	
	if ((strncmp("/dev/dlci.h5-baseband.3", path, 23) == 0) ||
		(strncmp("/dev/dlci.spi-baseband.3", path, 24) == 0)) {
		
		struct sockaddr_un saun;
		fd = socket(AF_UNIX, SOCK_STREAM, 0);	
		saun.sun_family = AF_UNIX;
		strcpy(saun.sun_path, FD3);
		int len = offsetof(struct sockaddr_un, sun_path) + strlen(FD3);
		if (connect(fd, &saun, len) < 0) {
			//printf("connect fd3 error\n");
			log("connect fd3 error\n", 0, 0);
		}
		else {
			sprintf(l, "connected fd3 (baseband.3) fd=%d orig_flags=%x\n", fd, flags);
			log(l, 0, 0);
			fd3 = fd;
		}
	}
	else if	((strncmp("/dev/dlci.h5-baseband.4", path, 23) == 0) || 
		(strncmp("/dev/dlci.spi-baseband.4", path, 24) == 0)) {

		struct sockaddr_un saun;
		fd = socket(AF_UNIX, SOCK_STREAM, 0);	
		saun.sun_family = AF_UNIX;
		strcpy(saun.sun_path, FD4);
		int len = offsetof(struct sockaddr_un, sun_path) + strlen(FD4);
		if (connect(fd, &saun, len) < 0) {
			//printf("connect fd4 error\n");
			log("connect fd4 error\n",0,0);
		}
		else {
			sprintf(l, "connected fd4 (baseband.4) fd=%d orig_flags=%x\n", fd, flags);
			log(l, 0, 0);
			fd4 = fd;
		}
	}
	else { 
		if (flags & O_CREAT) {
			va_list arg_list;
			mode_t mode;
			va_start(arg_list, flags);
			mode = va_arg(arg_list, int);
			va_end(arg_list);
			fd = real_open(path, flags, mode);
		}
		else {
			fd = real_open(path, flags);
		}

		sprintf(l, "%d = open(%s)\n", fd, path);
		log(l, 0, 0);
	}

	return fd;
}
